// Colour force simulation
#include <graphics.c>
#include <lset.c>
#include <rnd/frnd.c>
#include <rnd/Max.c>
#include <rnd/normalvariate.c>

typedef struct {double x,y,xs,ys; int c;} Quark;

const int Quark_colours=3; 

void Quarks_advanceplot(LSet *qs)
{
	#ifdef DECORATIVEPARTICLES
	static GLuint qtex=0;
	if (!qtex) qtex=quarktex(32);
	#endif
	unsigned col[3]={0xFF0000,0x00FF00,0x0000FF};
	int n,i; Quark *q; LSet_hookup(qs,(void **)&q);
	float zoom=1;//exp(3.0*((float)my/yres-0.5));
	static double ot=-1; if (ot<0) ot=eithertime();
	double dx,dy,dd,f,ka,kr,
		t=eithertime(),dt=10.0*(t-ot); ot=t;
	for (n=qs->m-1;n>=0;n--) for (i=n-1;i>=0;i--)
	{
		dx=q[i].x-q[n].x; dy=q[i].y-q[n].y;
		dd=dx*dx+dy*dy; ka=1.0/sqrt(dd); kr=1.0/(dd*dd);
		f=1.0/Quark_colours-(q[n].c==q[i].c);
		if (ka<0.02) ka*=ka/0.02;
		q[n].xs+=dt*dx*(f*ka-kr);
		q[n].ys+=dt*dy*(f*ka-kr);
		q[i].xs-=dt*dx*(f*ka-kr);
		q[i].ys-=dt*dy*(f*ka-kr);
	}
	double cool=exp(-0.01*dt);
	#ifdef DECORATIVEPARTICLES
	glEnable(GL_TEXTURE_2D); glBindTexture(GL_TEXTURE_2D,qtex);
	glEnable(GL_BLEND); glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
	glBegin(GL_QUADS);
	#endif
	unsigned c; float r;
	for (n=qs->m-1;n>=0;n--)
	{
		q[n].x+=q[n].xs*dt;
		q[n].y+=q[n].ys*dt;
		//q[n].xs*=cool; q[n].ys*=cool;
		if (fabs(q[n].x)>xres/2 || fabs(q[n].y)>yres/2) {LSet_remove(qs,n); continue;}
		c=col[q[n].c]; r=Max(2,0.5*zoom);
		#ifdef DECORATIVEPARTICLES
		glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
		texquad(xres/2+q[n].x*zoom,yres/2-q[n].y*zoom,r*14);
		glColor3f(1,1,1);
		texquad(xres/2+q[n].x*zoom,yres/2-q[n].y*zoom,r*7);
		#else				
		circlefill(xres/2+q[n].x*zoom,yres/2-q[n].y*zoom,r,c);
		#endif
	}
	#ifdef DECORATIVEPARTICLES
	glEnd();
	glDisable(GL_TEXTURE_2D); glDisable(GL_BLEND);
	#endif
}

void Quarks_addbaryons(LSet *qs,const int m)
{
	int n; Quark nq;
	for (n=m*Quark_colours;n>0;n--)
	{
		twonormalvariates(&nq.x,&nq.y);
		nq.x*=10; nq.y*=10;
		nq.xs=0; nq.ys=0;
		nq.c=n%Quark_colours;
		LSet_add(qs,&nq);
	}
}

/*// Example usage
WINMAIN
{
	GRAPHICS(0,0,"Colour force simulation [Stephen Brooks 2014]");
	LSet qs=LSet(Quark);
	LOOP
	{
		if (KEY(VK_SPACE)) Quarks_addbaryons(&qs,1);
		Quarks_advanceplot(&qs);
		ccl();
	}
}
*/

#define DECORATIVEQUARKS
