#ifndef INCLUDED_SBCLIB_GRAPHICS_OPENGL
// OpenGL needs a different "graphics layer 2" because writing bytes to the framebuffer is no longer fastest
#define SBCLIB_GRAPHICS
#define SBCLIB_GRAPHICS_OPENGL
#include <math.h>
#include <rnd/Mini.c>
#include <rnd/Maxi.c>
#ifdef SBCLIB_RPI
#include <pistack/gfx_glshim.c>
#else
#include <gfx_opengl.c> // Link opengl32.lib for Windows; some other things for Linux
#include <GL/glext.h> // Needed for cubemaps, download from http://www.opengl.org/registry/api/glext.h
#endif

char popchar(void)
{
	char ret=*gotch;
	if (ret) for (int n=0;gotch[n];n++) gotch[n]=gotch[n+1];
	return ret;
}

char gl_ortho=1;
#define GRAPHICS3D(X,Y,T) gl_ortho=0; GRAPHICS(X,Y,T)
#define glIsSupported(X) (!!strstr((char *)glGetString(GL_EXTENSIONS),X)) // Warning: this is fairly slow (and not substring-proof)

#define GFX_white 0xFFFFFF
#define RGB666(R,G,B) ((int)(R)*0x330000+(int)(G)*0x3300+(int)(B)*0x33)
#define RGB555(R,G,B) (((int)(R)*0xFF+3)/4*0x10000+((int)(G)*0xFF+3)/4*0x100+((int)(B)*0xFF+3)/4)

void setzzonly(const GLfloat zn,const GLfloat zf)
{ // Set the depth range without changing FOV etc.
	GLfloat m[16]; // GL stores matrices column-first
	glGetFloatv(GL_PROJECTION_MATRIX,m);
	m[10]=(zn+zf)/(zn-zf); // Third row produces numerator for depth buffer ordinate d=(Az+B)/-z, d(-zn)=1, d(-zf)=-1
	m[14]=2.0*zn*zf/(zn-zf);
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(m);
	glMatrixMode(GL_MODELVIEW);
}

void set43fovonly(const GLfloat degs)
{ // Set the FOV without changing depth range
	GLfloat m[16]; // GL stores matrices column-first
	glGetFloatv(GL_PROJECTION_MATRIX,m);
	GLfloat xtan=tan(degs degrees*0.5),
		ytan43=xtan*0.75,
		ytan=xtan*yres/xres;
	if (ytan<ytan43) {ytan=ytan43; xtan=ytan*xres/yres;}
	m[0]=1.0/xtan; m[5]=1.0/ytan;
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(m);
	glMatrixMode(GL_MODELVIEW);
}

GLfloat getfov(void)
{
	GLfloat m[16];
	glGetFloatv(GL_PROJECTION_MATRIX,m);
	GLfloat xtan=1.0/m[0];
	return atan(xtan) radians*2;
}

GLfloat get43fov(void)
{
	GLfloat m[16];
	glGetFloatv(GL_PROJECTION_MATRIX,m);
	GLfloat xtan=1.0/m[0];
	if ((GLfloat)yres/xres<0.75) {GLfloat ytan=xtan*yres/xres,ytan43=ytan; xtan=ytan43/0.75;}
	return atan(xtan) radians*2;
}

void setfovzz(const GLfloat degs,const GLfloat zn,const GLfloat zf)
{ // Write my own that does it in X
	glMatrixMode(GL_PROJECTION);
	GLfloat m[16]; // GL stores matrices column-first
	for (int i=0;i<16;i++) m[i]=0;
	GLfloat xtan=tan(degs degrees*0.5),ytan=xtan*yres/xres;
	m[0]=1.0/xtan; m[5]=1.0/ytan;
	m[11]=-1; // This in the fourth row is because Z is out of the screen in OpenGL, thus -z is the denominator for proper perspective
	m[10]=(zn+zf)/(zn-zf); // Third row produces numerator for depth buffer ordinate d=(Az+B)/-z, d(-zn)=1, d(-zf)=-1
	m[14]=2.0*zn*zf/(zn-zf);
	glLoadMatrixf(m);
	glMatrixMode(GL_MODELVIEW);
}

void set43fovzz(const GLfloat degs,const GLfloat zn,const GLfloat zf)
{ // Guarantees the 4:3 field will be visible on all monitor shapes
	glMatrixMode(GL_PROJECTION);
	GLfloat m[16]; // GL stores matrices column-first
	for (int i=0;i<16;i++) m[i]=0;
	GLfloat xtan=tan(degs degrees*0.5),
		ytan43=xtan*0.75,
		ytan=xtan*yres/xres;
	if (ytan<ytan43) {ytan=ytan43; xtan=ytan*xres/yres;}
	m[0]=1.0/xtan; m[5]=1.0/ytan;
	m[11]=-1; // This in the fourth row is because Z is out of the screen in OpenGL, thus -z is the denominator for proper perspective
	m[10]=(zn+zf)/(zn-zf); // Third row produces numerator for depth buffer ordinate d=(Az+B)/-z, d(-zn)=1, d(-zf)=-1
	m[14]=2.0*zn*zf/(zn-zf);
	glLoadMatrixf(m);
	glMatrixMode(GL_MODELVIEW);
}

#define setfov(A) setfovzz(A,0.1f,1000.0f)
#define set43fov(A) set43fovzz(A,0.1f,1000.0f)

void ReSizeGLScene(const int xs,int ys)
{
	if (ys<1) ys=1; // Prevent divide by zero later on...
	xres=xs; yres=ys;
	glViewport(0,0,xs,ys);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	if (gl_ortho) glOrtho(0,xs,ys,0,ys/2,-ys/2);
	else set43fov(90);
	glMatrixMode(GL_MODELVIEW);
}

void InitGL(const int xs,const int ys)	// This Will Be Called Right After The GL Window Is Created
{
	glClearColor(0.0f,0.0f,0.0f,0.0f);		// This Will Clear The Background Color To Black
	glClearDepth(1.0);							// Enables Clearing Of The Depth Buffer
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT); // Clear The Screen And The Depth Buffer
	glDepthFunc(GL_LEQUAL);						// The Type Of Depth Test To Do
	glEnable(GL_DEPTH_TEST);					// Enables Depth Testing
	glShadeModel(GL_SMOOTH);					// Enables Smooth Color Shading (linear interpolation across polys and lines)
	glEnable(GL_COLOR_MATERIAL); // So you still get colours with lighting enabled
	glEnable(GL_TEXTURE_CUBE_MAP_SEAMLESS); // Stops ugly seams when interpolating between cubemap faces
	ReSizeGLScene(xs,ys);
	glLoadIdentity(); // Set the GL_MODELVIEW matrix
}

float glversion(void)
{
	return atof((char *)glGetString(GL_VERSION));
}

void ccl(const unsigned c=0)
{ // Use glClearColor glClearDepth to set how this clearing works
	glFinish(); // Makes vsync work better
	#ifdef __linux__
	//SDL_GL_SwapBuffers();
    SDL_GL_SwapWindow(sdl_window);
	#else
	SwapBuffers(hDC);
	#endif
	static unsigned oc=0x1000000;
	if (c!=oc) {glClearColor((float)(c>>16)/255.0,(float)((c>>8)&0xFF)/255.0,(float)(c&0xFF)/255.0,0); oc=c;}
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT); // Clear The Screen And The Depth Buffer
	glLoadIdentity(); // Reset The View
}

void update(void)
{
	glFinish();
	for (int n=0;n<2;n++) // New way, works on 10" laptop
	{
		#ifdef __linux__
		//SDL_GL_SwapBuffers();
	    SDL_GL_SwapWindow(sdl_window);
		#else
		SwapBuffers(hDC);
		#endif
	}
	glDrawBuffer(GL_FRONT); // Should be this anyway (messes up on 10" laptop)
	glLoadIdentity(); // Reset The View
}

void clearrect(const int ox,const int oy,const int xs,const int ys)
{ // Clears both the colour and depth buffers in a rectangular region
	glEnable(GL_SCISSOR_TEST); glScissor(ox,oy,xs,ys);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glDisable(GL_SCISSOR_TEST);
}

void cl(const unsigned c=0)
{
	ccl(c); ccl(c);
}

void plotdefaultmouse(const float b=1)
{ // Only works in ortho view
	GLboolean oldlight=glIsEnabled(GL_LIGHTING); glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	glColor3f(b,b,b);
	glVertex2f(mx+0.1,my+0.1); glVertex2f(mx+xres/200,my+0.1);
	glVertex2f(mx+0.1,my+0.1); glVertex2f(mx+0.1,my+xres/200);
	glVertex2f(mx+0.1,my+0.1); glVertex2f(mx+xres/100,my+xres/100);
	glEnd();
	if (oldlight) glEnable(GL_LIGHTING);
}

int contrastshade(const unsigned c)
{ // 0 for black, 1 for white
	if ((c&0xFF)*3+((c>>8)&0xFF)*6+((c>>16)&0xFF)*2>=0.5*0xFF*11) return 0;
	else return 1;
}

void plotvisiblemouse(void)
{ // Only works in ortho view
	unsigned c;
	glReadPixels(Mini(xres-1,mx+xres/200),yres-1-Mini(my+yres/200,yres-1),1,1,GL_RGB,GL_UNSIGNED_BYTE,&c);
	plotdefaultmouse(contrastshade(c));
}

int screenshotrect_pitch(const int xs) {return (xs*3+3)/4*4;}

unsigned char *screenshotrect_data(const int ox,const int oy,const int xs,const int ys)
{
	unsigned char *d=(unsigned char *)malloc(ys*screenshotrect_pitch(xs));
	glReadPixels(ox,yres-(oy+ys),xs,ys,GL_BGR,GL_UNSIGNED_BYTE,d);
	return d;
}

#include <rnd/setextension.c>
#include <rnd/fileexists.c>
#include <rnd/sjbdeletefile.c>
void screenshotrect_save(const unsigned char *d,const int xs,const int ys,const char *filename,const int png=1)
{ // filename has the ".bmp" extension in all these functions
	FILE *out=fopen(filename,"wb");
	int pitch=screenshotrect_pitch(xs),
		mhead=18+2*sizeof(unsigned)+28,mimg=ys*pitch,m=mhead+mimg;
	unsigned char seg1[18]={0x42,0x4D,0x46,0x0E,3,0,0,0,0,0,0x36,0,0,0,0x28,0,0,0},
		seg2[28]={1,0,0x18,0,0,0,0,0,0x10,0x0E,3,0,0x12,0x0B,0,0,0x12,0x0B,0,0,0,0,0,0,0,0,0,0};
	fwrite(seg1,1,2,out);
	fwrite(&m,1,sizeof(unsigned),out);
	fwrite(seg1+6,1,12,out);
	fwrite(&xs,sizeof(int),1,out);
	fwrite(&ys,sizeof(int),1,out);
	fwrite(seg2,1,28,out);
	fwrite(d,ys*pitch,1,out);
	fclose(out);
	if (png)
	{
		char cmd[300];
		//sprintf(cmd,"bmp2png -9 -E \"%s\"",filename);
		char *filepng=setextension(filename,"png"); sprintf(cmd,"magick -quality 95 \"%s\" \"%s\"",filename,filepng);
		system(cmd);
		if (fileexists(filepng)) sjbdeletefile(filename);
		free(filepng);
	}
}

void screenshotrect(const int ox,const int oy,const int xs,const int ys,const char *filename,const int png=1)
{ // filename has the ".bmp" extension in all these functions
	unsigned char *d=screenshotrect_data(ox,oy,xs,ys);
	screenshotrect_save(d,xs,ys,filename,png);
	free(d);
}

void screenshot(const char *file,const int png=1) {screenshotrect(0,0,xres,yres,file,png);}

#include <rnd/eithertime.c>
void screenshotauto(const char *base="screenshot")
{
	char file[500];
	sprintf(file,"%s_%05d.bmp",base,((int)(eithertime()/0.3))%100000);
	screenshot(file);
}

void setcolour(const int c)
{ // Emulates my 8-bit palette
	if (c<16)
	{
		if (c==8) glColor3f(0.25,0.25,0.25);
		else {float b=0.5+0.5*(c>8); glColor3f((c&4)?b:0,(c&2)?b:0,(c&1)?b:0);}
	}
	else if (c<40) {float b=(float)(c-16)/23; glColor3f(b,b,b);}
	else glColor3f(0.2*((c-40)%6),0.2*((c-40)/6%6),0.2*((c-40)/36));
}

void smoothing_on(const int additive=0,const float sz=1.5)
{
	glEnable(GL_BLEND); // These four lines enable line smoothing, only glHint is optional
	glBlendFunc(GL_SRC_ALPHA,additive?GL_ONE:GL_ONE_MINUS_SRC_ALPHA); // Was GL_SRC_ALPHA_SATURATE, dunno why
	glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
	glEnable(GL_LINE_SMOOTH); glLineWidth(sz); // Line width 1 causes roping on gamma~2.2 displays
	glHint(GL_POINT_SMOOTH_HINT,GL_NICEST);
	glEnable(GL_POINT_SMOOTH); glPointSize(sz);
}

void smoothing_off(void)
{
	glDisable(GL_BLEND);
	glDisable(GL_LINE_SMOOTH); glLineWidth(1);
	glDisable(GL_POINT_SMOOTH); glPointSize(1);
}

void ortho_on(void)
{
	glPushMatrix(); // Save current model matrix state
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix(); // Save current projection matrix state
	glLoadIdentity();
	int v[4]; glGetIntegerv(GL_VIEWPORT,v);
	glOrtho(v[0],v[0]+v[2],yres-v[1],yres-(v[1]+v[3]),yres/2,-yres/2);
	glMatrixMode(GL_MODELVIEW);
	glDisable(GL_DEPTH_TEST); glDepthMask(0);
}

void ortho_off(void)
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glEnable(GL_DEPTH_TEST); glDepthMask(1);
}

double ortho_pixelsize(void)
{ // Returns 1 screen pixel's logical size (erring on the average side) when in an orthogonal projection
	int v[4]; glGetIntegerv(GL_VIEWPORT,v);
	GLdouble pm[16],vm[16],m[16];
	glGetDoublev(GL_PROJECTION_MATRIX,pm);
	glGetDoublev(GL_MODELVIEW_MATRIX,vm);
	int i,j,k; for (i=0;i<4;i++) for (j=0;j<4;j++)
	{
		m[i+j*4]=0;
		for (k=0;k<4;k++) m[i+j*4]+=pm[i+k*4]*vm[k+j*4];
	}
	for (i=0;i<4;i++)
	{
		m[i+0*4]*=0.5*v[2];
		m[i+1*4]*=0.5*v[3];
	}
	return 1.0/sqrt(fabs(m[0]*m[5]-m[1]*m[4])); // sqrt(det(screen->logical))
}

double radian_pixelsize(void)
{ // Radians (in X) per pixel at centre of screen in 3D mode
	GLdouble m[16];
	glGetDoublev(GL_PROJECTION_MATRIX,m);
	return 2.0/(m[0]*xres);
}

void onscreenpos_inner(const double x,const double y,const double z,double *px,double *py,
	const GLdouble pm[],const GLdouble vm[],const int v[])
{
	int i; GLdouble a[4],b[4];
	a[0]=x; a[1]=y; a[2]=z; a[3]=1;
	for (i=0;i<4;i++) b[i]=vm[i]*a[0]+vm[i+4]*a[1]+vm[i+8]*a[2]+vm[i+12]*a[3];
	for (i=0;i<4;i++) a[i]=pm[i]*b[0]+pm[i+4]*b[1]+pm[i+8]*b[2]+pm[i+12]*b[3];
	*px=v[0]+(a[0]/a[3]*0.5+0.5)*v[2];
	*py=v[1]+(a[1]/a[3]*-0.5+0.5)*v[3];
}

void onscreenpos(const double x,const double y,const double z,double *px,double *py)
{ // Translates 3D to my standard 2D coordinates (as used in bfwrite, ortho_on mode etc.)
	int v[4]; glGetIntegerv(GL_VIEWPORT,v);
	GLdouble pm[16],vm[16];
	glGetDoublev(GL_PROJECTION_MATRIX,pm);
	glGetDoublev(GL_MODELVIEW_MATRIX,vm);
	onscreenpos_inner(x,y,z,px,py,pm,vm,v);
}

GLdouble onscreenpos_pm[16],onscreenpos_vm[16]; int onscreenpos_v[4];

void onscreenpos_fast_init(void)
{
	int v[4]; glGetIntegerv(GL_VIEWPORT,v);
	GLdouble pm[16],vm[16];
	glGetDoublev(GL_PROJECTION_MATRIX,pm);
	glGetDoublev(GL_MODELVIEW_MATRIX,vm);
}

void onscreenpos_fast(const double x,const double y,const double z,double *px,double *py)
{ // Might be faster but also can be used between glBegin/End
	onscreenpos_inner(x,y,z,px,py,onscreenpos_pm,onscreenpos_vm,onscreenpos_v);
}

// Variables for use with clipping rectangles
typedef struct {int x1,y1,x2,y2;} GFX_Clipper;

void GFX_Clipper_setcliprect(const GFX_Clipper c)
{
	glEnable(GL_SCISSOR_TEST);
	glScissor(c.x1,yres-c.y2,c.x2-c.x1,c.y2-c.y1);
}

GFX_Clipper GFX_Clipper_getcliprect(void)
{
	GFX_Clipper ret; int v[4];
	glGetIntegerv(GL_SCISSOR_BOX,v);
	ret.x1=v[0];
	ret.x2=ret.x1+v[2];
	ret.y2=yres-v[1];
	ret.y1=ret.y2-v[3];
	return ret;
}

#define GFX_clipM 200 // Maximum nesting
int GFX_clipm=0; // Current nesting level
GFX_Clipper GFX_clip[GFX_clipM];

void setcliprect(int x1,int y1,int x2,int y2)
{
	if (GFX_clipm>=GFX_clipM) return;
	GFX_Clipper c;
	if (GFX_clipm>0)
	{
		GFX_Clipper oc=GFX_clip[GFX_clipm-1];
		c.x1=Mini(Maxi(x1,oc.x1),oc.x2); c.x2=Mini(Maxi(x2,oc.x1),oc.x2);
		c.y1=Mini(Maxi(y1,oc.y1),oc.y2); c.y2=Mini(Maxi(y2,oc.y1),oc.y2);
	}
	else {c.x1=x1; c.x2=x2; c.y1=y1; c.y2=y2;}
	GFX_clip[GFX_clipm]=c; GFX_clipm++;
	GFX_Clipper_setcliprect(c);
}

void unsetcliprect(void)
{
	if (GFX_clipm<=0) return;
	GFX_clipm--;
	if (GFX_clipm>0) GFX_Clipper_setcliprect(GFX_clip[GFX_clipm-1]);
	else glDisable(GL_SCISSOR_TEST);
}

void glColor3(const unsigned c)
{ // Uses "HTML" byte ordering
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
}

// 2D graphics functions
void putpix(const double x,const double y,const unsigned c)
{ // I pity the fool who uses this in a loop
	glBegin(GL_POINTS);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y);
	glEnd();
}

void putpix_inner(const double x,const double y,const unsigned c)
{ // glBegin(GL_POINTS)...
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y);
}

void line(const double x0,const double y0,const double x1,const double y1,const unsigned c)
{ // Not really recommended, more for 2D compatibility
	GLboolean oldlight=glIsEnabled(GL_LIGHTING); glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x0,y0); glVertex2d(x1,y1);
	glEnd();
	if (oldlight) glEnable(GL_LIGHTING);
}

void line_inner(const double x0,const double y0,const double x1,const double y1,const unsigned c)
{ // glBegin(GL_LINES)...
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x0,y0); glVertex2d(x1,y1);
}

void rect(const double x,const double y,const double xs,const double ys,const unsigned c)
{ // Probably slow, use rect_inner if possible
	glBegin(GL_QUADS);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y); glVertex2d(x+xs,y); glVertex2d(x+xs,y+ys); glVertex2d(x,y+ys);
	glEnd();
}

void rect_inner(const double x,const double y,const double xs,const double ys,const unsigned c)
{ // glBegin(GL_QUADS)...
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y); glVertex2d(x+xs,y); glVertex2d(x+xs,y+ys); glVertex2d(x,y+ys);
}

void box(const double x,const double y,const double xs,const double ys,const unsigned c)
{ // Probably slow, use box_inner (which doesn't exist yet) if possible
	glBegin(GL_LINE_LOOP);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y); glVertex2d(x+xs,y); glVertex2d(x+xs,y+ys); glVertex2d(x,y+ys);
	glEnd();
}

void boxa(const double x0,const double y0,const double x1,const double y1,const unsigned c)
{ // Probably slow, use boxa_inner (which doesn't exist yet) if possible
	glBegin(GL_LINE_LOOP);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x0,y0); glVertex2d(x1,y0); glVertex2d(x1,y1); glVertex2d(x0,y1);
	glEnd();
}

int circlecliptest=1;

void circlefill(const double x,const double y,const double r,const unsigned c)
{ // No need to do "_inner" of this, need to start new TRIANGLE_FAN anyway
	if (circlecliptest) if (x-r>xres || x+r<0 || y-r>yres || y+r<0) return;
	int n,dot_sides=(int)(5.0*sqrt(r/ortho_pixelsize())); if (dot_sides<1) dot_sides=1;
	double step=M_TWOPI/dot_sides,th;
	glBegin(GL_TRIANGLE_FAN);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glVertex2d(x,y);
	for (n=dot_sides,th=0;n>=0;n--,th+=step)
		glVertex2d(x+r*sin(th),y+r*cos(th));
	glEnd();
}

void circle(const double x,const double y,const double r,const unsigned c)
{ // Neither of these are particularly fast right now
	if (circlecliptest)
	{
		if (x-r>xres || x+r<0 || y-r>yres || y+r<0) return;
		double ss=(xres/2-x)*(xres/2-x)+(yres/2-y)*(yres/2-y),w=0.5*(xres+yres);
		if (r>w) if (fabs(ss-r*r-w*w)>r*w*2) return;
	}
	int n,dot_sides=(int)(5.0*sqrt(r/ortho_pixelsize())); if (dot_sides<1) dot_sides=1;
	double step=M_TWOPI/dot_sides,th;
	glBegin(GL_LINE_LOOP);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	for (n=dot_sides,th=0;n>0;n--,th+=step)
		glVertex2d(x+r*sin(th),y+r*cos(th));
	glEnd();
}

void ellipse(const double x,const double y,const double rx,const double ry,const unsigned c)
{ // Neither of these are particularly fast right now (no cliptest in this one either)
	int n,dot_sides=(int)(5.0*sqrt(max(rx,ry)/ortho_pixelsize())); if (dot_sides<1) dot_sides=1;
	double step=M_TWOPI/dot_sides,th;
	glBegin(GL_LINE_LOOP);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	for (n=dot_sides,th=0;n>0;n--,th+=step)
		glVertex2d(x+rx*sin(th),y+ry*cos(th));
	glEnd();
}

void circlearc(const double x,const double y,const double r,const double th1,const double th2,const unsigned c)
{ // Neither of these are particularly fast right now (th measured clockwise)
	if (circlecliptest)
	{
		if (x-r>xres || x+r<0 || y-r>yres || y+r<0) return;
		double ss=(xres/2-x)*(xres/2-x)+(yres/2-y)*(yres/2-y),w=0.5*(xres+yres);
		if (r>w) if (fabs(ss-r*r-w*w)>r*w*2) return;
	}
	int n,dot_sides=(int)(0.8*fabs(th2-th1)*sqrt(fabs(r)/ortho_pixelsize())); if (dot_sides<1) dot_sides=1;
	double step=(th2-th1)/dot_sides,th;
	glBegin(GL_LINE_STRIP);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	for (n=dot_sides,th=th1;n>=0;n--,th+=step)
		glVertex2d(x+r*sin(th),y-r*cos(th));
	glEnd();
}

void circular(const double x,const double y,const double r1,const double r2,const double th1,const double th2,const unsigned c,const unsigned char a=0xFF)
{ // theta should be as on a clock for inverted Y-coordinate systems
	if (circlecliptest) if (x-r2>xres || x+r2<0 || y-r2>yres || y+r2<0) return;
	int n,dot_sides=(int)(0.8*(th2-th1)*sqrt(r2/ortho_pixelsize())); if (dot_sides<1) dot_sides=1;
	double step=(th2-th1)/dot_sides,th;
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,a);
	if (r1==0)
	{
		glBegin(GL_TRIANGLE_FAN);
		glVertex2d(x,y);
		for (n=dot_sides,th=th1;n>=0;n--,th+=step)
			glVertex2d(x+r2*sin(th),y-r2*cos(th));
		glEnd();
	}
	else
	{
		double sth,cth;
		glBegin(GL_QUAD_STRIP);
		for (n=dot_sides,th=th1;n>=0;n--,th+=step)
		{
			glVertex2d(x+r1*(sth=sin(th)),y-r1*(cth=cos(th)));
			glVertex2d(x+r2*sth,y-r2*cth);
		}
		glEnd();
	}
}

void glBlendFuncSeparate(GLenum srcRGB,GLenum dstRGB,GLenum srcAlpha,GLenum dstAlpha)
{
	void static (APIENTRY *w_glBlendFuncSeparate)(GLenum srcRGB,GLenum dstRGB,GLenum srcAlpha,GLenum dstAlpha)=NULL;
	if (!w_glBlendFuncSeparate) *(void **)&w_glBlendFuncSeparate=sjbglGetProcAddress("glBlendFuncSeparate");
	w_glBlendFuncSeparate(srcRGB,dstRGB,srcAlpha,dstAlpha);
}

void circulars(const double x,const double y,const double r1,const double r2,
	const double th1,const double th2,const unsigned c)
{ // 's' is for "smooth" (these functions mess up smoothing_on)
	if (circlecliptest) if (x-r2>xres || x+r2<0 || y-r2>yres || y+r2<0) return;
	glEnable(GL_BLEND); //glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glBlendFuncSeparate(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA,GL_ONE,GL_ONE);
	double pix=ortho_pixelsize();
	int n,dot_sides=(int)(0.8*(th2-th1)*sqrt(r2/pix)); if (dot_sides<1) dot_sides=1;
	double step=(th2-th1)/dot_sides,th,sth,cth;
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glBegin(GL_QUAD_STRIP);
	for (n=dot_sides,th=th1;n>=0;n--,th+=step)
	{
		sth=sin(th); cth=cos(th);
		glVertex2d(x+r1*sth,y-r1*cth);
		glVertex2d(x+r2*sth,y-r2*cth);
	}
	glEnd();
	glBegin(GL_QUAD_STRIP); // Outer edge
	for (n=dot_sides,th=th1;n>=0;n--,th+=step)
	{
		sth=sin(th); cth=cos(th);
		glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF);
		glVertex2d(x+r2*sth,y-r2*cth);
		glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
		glVertex2d(x+(r2+pix)*sth,y-(r2+pix)*cth);
	}
	glEnd();
	if (r1>0)
	{
		glBegin(GL_QUAD_STRIP); // Inner edge
		for (n=dot_sides,th=th1;n>=0;n--,th+=step)
		{
			sth=sin(th); cth=cos(th);
			glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF);
			glVertex2d(x+r1*sth,y-r1*cth);
			glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
			glVertex2d(x+(r1-pix)*sth,y-(r1-pix)*cth);
		}
		glEnd();
	}
	if (th2<th1+M_TWOPI)
	{
		double nx,ny;
		glBegin(GL_QUADS);
		cth=cos(th1); sth=sin(th1);
		nx=pix*-cth; ny=pix*sth;
		glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF);
		glVertex2d(x+r1*sth,y-r1*cth);
		glVertex2d(x+r2*sth,y-r2*cth);
		glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
		glVertex2d(x+r2*sth+nx,y-r2*cth-ny);
		glVertex2d(x+r1*sth+nx,y-r1*cth-ny);
		cth=cos(th2); sth=sin(th2);
		nx=pix*cth; ny=pix*-sth;
		glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF);
		glVertex2d(x+r1*sth,y-r1*cth);
		glVertex2d(x+r2*sth,y-r2*cth);
		glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
		glVertex2d(x+r2*sth+nx,y-r2*cth-ny);
		glVertex2d(x+r1*sth+nx,y-r1*cth-ny);
		glEnd();
	}
	glDisable(GL_BLEND);
}

#define circlefills(x,y,r,c) circulars(x,y,0,r,0,M_TWOPI,c)

void rectangles(const double x,const double y,const double xs,const double ys,
	const double ca,const double sa,const unsigned c) // ca=1, sa=0 is unrotated
{ // "angle" means rotated; 's' means "smooth"
	glEnable(GL_BLEND); //glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glBlendFuncSeparate(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA,GL_ONE,GL_ONE);
	double pix=ortho_pixelsize(),x1,y1,x2,y2,x3,y3;
	glBegin(GL_QUADS);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF);
	glVertex2d(x,y); // Main quad ABCD
	glVertex2d(x1=x+xs*ca,y1=y+xs*sa);
	glVertex2d(x2=x+xs*ca-ys*sa,y2=y+xs*sa+ys*ca);
	glVertex2d(x3=x-ys*sa,y3=y+ys*ca);
	glVertex2d(x,y); // Side AB
	glVertex2d(x1,y1);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
	glVertex2d(x1+pix*sa,y1-pix*ca);
	glVertex2d(x+pix*sa,y-pix*ca);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF); // Side BC
	glVertex2d(x1,y1);
	glVertex2d(x2,y2);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
	glVertex2d(x2+pix*ca,y2+pix*sa);
	glVertex2d(x1+pix*ca,y1+pix*sa);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF); // Side CD
	glVertex2d(x2,y2);
	glVertex2d(x3,y3);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
	glVertex2d(x3-pix*sa,y3+pix*ca);
	glVertex2d(x2-pix*sa,y2+pix*ca);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF); // Side DA
	glVertex2d(x3,y3);
	glVertex2d(x,y);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
	glVertex2d(x-pix*ca,y-pix*sa);
	glVertex2d(x3-pix*ca,y3-pix*sa);
	glEnd();
	glDisable(GL_BLEND);
}

void triangles(const double x1,const double y1,const double x2,const double y2,
	const double x3,const double y3,const unsigned c)
{ // 's' means "smooth"
	glBegin(GL_TRIANGLES);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF);
	glVertex2d(x1,y1); glVertex2d(x2,y2); glVertex2d(x3,y3); // Main triangle ABC
	glEnd();
	glEnable(GL_BLEND); //glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glBlendFuncSeparate(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA,GL_ONE,GL_ONE);
	double pix=ortho_pixelsize(),ux,uy,k;
	if ((x2-x1)*(y3-y1)<(x3-x1)*(y2-y1)) pix*=-1; // Determinant thingy to get winding direction right
	glBegin(GL_QUADS);
	ux=x2-x1; uy=y2-y1; k=pix/sqrt(ux*ux+uy*uy); ux*=k; uy*=k; // Side AB
	glVertex2d(x1,y1);
	glVertex2d(x2,y2);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
	glVertex2d(x2+uy,y2-ux);
	glVertex2d(x1+uy,y1-ux);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF); // Side BC
	ux=x3-x2; uy=y3-y2; k=pix/sqrt(ux*ux+uy*uy); ux*=k; uy*=k;
	glVertex2d(x2,y2);
	glVertex2d(x3,y3);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
	glVertex2d(x3+uy,y3-ux);
	glVertex2d(x2+uy,y2-ux);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0xFF); // Side CA
	ux=x1-x3; uy=y1-y3; k=pix/sqrt(ux*ux+uy*uy); ux*=k; uy*=k;
	glVertex2d(x3,y3);
	glVertex2d(x1,y1);
	glColor4ub(c>>16,(c>>8)&0xFF,c&0xFF,0);
	glVertex2d(x1+uy,y1-ux);
	glVertex2d(x3+uy,y3-ux);
	glEnd();
	glDisable(GL_BLEND);
}

#include <rnd/sjbbuild2dmipmaps.c>

void glGenerateMipmap(GLenum target)
{
	static int supported=-1;
	if (supported<0) supported=glIsSupported("GL_SGIS_generate_mipmap");
	if (!supported) {sjbBuild2DMipmaps_existing(); return;}
	void static (APIENTRY *w_glGenerateMipmap)(GLenum target)=NULL;
	if (!w_glGenerateMipmap) *(void **)&w_glGenerateMipmap=sjbglGetProcAddress("glGenerateMipmap");
	w_glGenerateMipmap(target);
}

void glBlendEquation(GLenum mode)
{
	void static (APIENTRY *w_glBlendEquation)(GLenum mode)=NULL;
	if (!w_glBlendEquation) *(void **)&w_glBlendEquation=sjbglGetProcAddress("glBlendEquation");
	w_glBlendEquation(mode);
}

#ifdef SBCLIB_BMP
#include <rnd/enclosingpoweroftwo.c>

GLuint overloaded bmptexture(bmp b,int *ptx/*=NULL*/,int *pty/*=NULL*/)
{ // Texture from 24-bit BMP
	GLuint ret;
	glGenTextures(1,&ret);
	glBindTexture(GL_TEXTURE_2D,ret); // Selects the current texture
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	if (b.xs==0) {b.xs=b.ys=1; b.d=(unsigned char *)calloc(b.cpp,1); b.d[0]=0x80;}
	int tx=enclosingpoweroftwo(b.xs),ty=enclosingpoweroftwo(b.ys);
	if (ptx) *ptx=tx; if (pty) *pty=ty;
	int pitch=(tx*b.cpp+3)/4*4; // Round up to 4-byte boundary
	unsigned char *d=(unsigned char *)calloc(ty*pitch,1),*br,*dr;
	int x,y;
	for (y=b.ys-1;y>=0;y--)
	{
		dr=d+(b.ys-1-y)*pitch; br=b.d+y*b.xs*b.cpp;
		if (b.cpp==3) for (x=b.xs-1;x>=0;x--)
			{dr[x*3]=br[x*3+2]; dr[x*3+1]=br[x*3+1]; dr[x*3+2]=br[x*3];}
		else if (b.cpp==4) for (x=b.xs-1;x>=0;x--)
			{dr[x*4]=br[x*4+2]; dr[x*4+1]=br[x*4+1]; dr[x*4+2]=br[x*4]; dr[x*4+3]=br[x*4+3];}
	}
	sjbBuild2DMipmaps(b.cpp,tx,ty,d);
	free(d);
	return ret;
}

GLuint overloaded bmptexture(bmp b) {return bmptexture(b,NULL,NULL);} // For LCC

GLuint overloaded bmptexture(const char *filename)
{
	bmp b=bmpload(filename);
	GLuint ret=bmptexture(b);
	bmpfree(b);
	return ret;
}

GLuint bmptextureresource(const unsigned char *data)
{
	bmp b=bmploaddata(data);
	GLuint ret=bmptexture(b);
	bmpfree(b);
	return ret;
}
#endif

GLuint screentexture(const int x,const int y,const int xs,const int ys,const int use_alpha=0,
	const GLuint update=0)
{
	GLuint ret;
	if (!update) glGenTextures(1,&ret); else ret=update;
	glBindTexture(GL_TEXTURE_2D,ret);
	if (!update)
	{
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	}
	if (update) glCopyTexSubImage2D(GL_TEXTURE_2D,0, 0,0, x,yres-(y+ys),xs,ys);
	else glCopyTexImage2D(GL_TEXTURE_2D,0,use_alpha?GL_RGBA8:GL_RGB8, x,yres-(y+ys),xs,ys, 0);
	glGenerateMipmap(GL_TEXTURE_2D);
	return ret;
}

void aniso(const GLenum textype=GL_TEXTURE_2D)
{
	static int supported=-1;
	if (supported<0) supported=glIsSupported("GL_EXT_texture_filter_anisotropic");
	if (!supported) return;
	static float maxaniso=0;
	//if (KEY(VK_A)) maxaniso=1; else maxaniso=0;
	if (!maxaniso)
	{
		glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT,&maxaniso);
		//FILE *out=fopen("aniso.log","at"); fprintf(out,"maxaniso = %lg\n",maxaniso); fclose(out);
	}
	glTexParameterf(textype,GL_TEXTURE_MAX_ANISOTROPY_EXT,maxaniso);
}

#ifndef __linux__
void vsync(const int i)
{ // NB: the Linux equivalent of this is the GLX_SGI_swap_control extension I think
	static int supported=-1;
	if (supported<0) supported=glIsSupported("WGL_EXT_swap_control");
	if (!supported) return;
	BOOL static (APIENTRY *wglSwapIntervalEXT)(int interval)=NULL; // APIENTRY resolves to __stdcall
	if (!wglSwapIntervalEXT) *(void **)&wglSwapIntervalEXT=(void *)wglGetProcAddress("wglSwapIntervalEXT");
	wglSwapIntervalEXT(i);
}
#endif

// 3D graphics functions
void trinormal_inner(
	const double x1,const double y1,const double z1,
	const double x2,const double y2,const double z2,
	const double x3,const double y3,const double z3,const unsigned c)
{ // Must be within glBegin(GL_TRIANGLES)...
	double nx=(y2-y1)*(z3-z1)-(y3-y1)*(z2-z1),
		ny=(z2-z1)*(x3-x1)-(z3-z1)*(x2-x1),
		nz=(x2-x1)*(y3-y1)-(x3-x1)*(y2-y1),
		k=1.0/sqrt(nx*nx+ny*ny+nz*nz);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glNormal3d(k*nx,k*ny,k*nz);
	glVertex3d(x1,y1,z1);
	glVertex3d(x2,y2,z2);
	glVertex3d(x3,y3,z3);
}

void quadnormal_inner(
	const double x1,const double y1,const double z1,
	const double x2,const double y2,const double z2,
	const double x3,const double y3,const double z3,
	const double x4,const double y4,const double z4,const unsigned c)
{ // Must be within glBegin(GL_QUADS) (1234 is circular order)
	double ax=0.5*(x2+x3-x1-x4),bx=0.5*(x4+x3-x1-x2),
		ay=0.5*(y2+y3-y1-y4),by=0.5*(y4+y3-y1-y2),
		az=0.5*(z2+z3-z1-z4),bz=0.5*(z4+z3-z1-z2),
		nx=ay*bz-by*az,
		ny=az*bx-bz*ax,
		nz=ax*by-bx*ay,
		k=1.0/sqrt(nx*nx+ny*ny+nz*nz);
	glColor3ub(c>>16,(c>>8)&0xFF,c&0xFF);
	glNormal3d(k*nx,k*ny,k*nz);
	glVertex3d(x1,y1,z1);
	glVertex3d(x2,y2,z2);
	glVertex3d(x3,y3,z3);
	glVertex3d(x4,y4,z4);
}

void sphere_inner(const double ox,const double oy,const double oz,const double r,const unsigned col,const int m=10)
{ // 4*m*m triangles, m is number of divisions in 180 degrees
	int n,i; double th,
		*c=(double *)malloc((m+1)*sizeof(double)),
		*s=(double *)malloc((m+1)*sizeof(double));
	for (n=m;n>=0;n--) {th=M_PI*n/m; c[n]=cos(th); s[n]=sin(th);}
	glColor3ub(col>>16,(col>>8)&0xFF,col&0xFF);
	for (n=m-1;n>=0;n--) for (i=m-1;i>=0;i--)
	{
		glNormal3d(c[n]*s[i],s[n]*s[i],c[i]);
		glVertex3d(ox+r*c[n]*s[i],oy+r*s[n]*s[i],oz+r*c[i]);
		glNormal3d(c[n]*s[i+1],s[n]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n]*s[i+1],oy+r*s[n]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n+1]*s[i+1],s[n+1]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n+1]*s[i+1],oy+r*s[n+1]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n+1]*s[i],s[n+1]*s[i],c[i]);
		glVertex3d(ox+r*c[n+1]*s[i],oy+r*s[n+1]*s[i],oz+r*c[i]);
		glNormal3d(c[n]*s[i],-s[n]*s[i],c[i]);
		glVertex3d(ox+r*c[n]*s[i],oy-r*s[n]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i],-s[n+1]*s[i],c[i]);
		glVertex3d(ox+r*c[n+1]*s[i],oy-r*s[n+1]*s[i],oz+r*c[i]);
		glNormal3d(c[n+1]*s[i+1],-s[n+1]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n+1]*s[i+1],oy-r*s[n+1]*s[i+1],oz+r*c[i+1]);
		glNormal3d(c[n]*s[i+1],-s[n]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n]*s[i+1],oy-r*s[n]*s[i+1],oz+r*c[i+1]);
	}
	free(c); free(s);
}

void sphere(const double ox,const double oy,const double oz,const double r,const unsigned col,const int m=10)
{
	glBegin(GL_QUADS);
	sphere_inner(ox,oy,oz,r,col,m);
	glEnd();
}

void spheretex_inner(const double ox,const double oy,const double oz,const double r,
	const unsigned col=GFX_white,const int m=10,const double texscale=1)
{ // Uses x=longitude[-180,180], y=latitude[-90,90]; currently set up to look right from *inside* the sphere
	int n,i; double th,
		*c=(double *)malloc((m+1)*sizeof(double)),
		*s=(double *)malloc((m+1)*sizeof(double)),
		im=(double)1.0/m,hm=(double)0.5/m;
	for (n=m;n>=0;n--) {th=M_PI*n/m; c[n]=cos(th); s[n]=sin(th);}
	glColor3ub(col>>16,(col>>8)&0xFF,col&0xFF);
	double tx0,tx1,ty0,ty1;
	for (n=m-1;n>=0;n--) for (i=m-1;i>=0;i--)
	{
		tx0=(1.0-n*hm)*texscale; tx1=(1.0-(n+1)*hm)*texscale;
		ty0=(1.0-i*im)*texscale; ty1=(1.0-(i+1)*im)*texscale;
		glTexCoord2d(tx0,ty0);
		glNormal3d(c[n]*s[i],s[n]*s[i],c[i]);
		glVertex3d(ox+r*c[n]*s[i],oy+r*s[n]*s[i],oz+r*c[i]);
		glTexCoord2d(tx1,ty0);
		glNormal3d(c[n+1]*s[i],s[n+1]*s[i],c[i]);
		glVertex3d(ox+r*c[n+1]*s[i],oy+r*s[n+1]*s[i],oz+r*c[i]);
		glTexCoord2d(tx1,ty1);
		glNormal3d(c[n+1]*s[i+1],s[n+1]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n+1]*s[i+1],oy+r*s[n+1]*s[i+1],oz+r*c[i+1]);
		glTexCoord2d(tx0,ty1);
		glNormal3d(c[n]*s[i+1],s[n]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n]*s[i+1],oy+r*s[n]*s[i+1],oz+r*c[i+1]);
		
		tx0=n*hm*texscale; tx1=(n+1)*hm*texscale;
		glTexCoord2d(tx0,ty0);
		glNormal3d(c[n]*s[i],-s[n]*s[i],c[i]);
		glVertex3d(ox+r*c[n]*s[i],oy-r*s[n]*s[i],oz+r*c[i]);
		glTexCoord2d(tx0,ty1);
		glNormal3d(c[n]*s[i+1],-s[n]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n]*s[i+1],oy-r*s[n]*s[i+1],oz+r*c[i+1]);
		glTexCoord2d(tx1,ty1);
		glNormal3d(c[n+1]*s[i+1],-s[n+1]*s[i+1],c[i+1]);
		glVertex3d(ox+r*c[n+1]*s[i+1],oy-r*s[n+1]*s[i+1],oz+r*c[i+1]);
		glTexCoord2d(tx1,ty0);
		glNormal3d(c[n+1]*s[i],-s[n+1]*s[i],c[i]);
		glVertex3d(ox+r*c[n+1]*s[i],oy-r*s[n+1]*s[i],oz+r*c[i]);
	}
	free(c); free(s);
}

void spheretex(const double ox,const double oy,const double oz,const double r,const GLuint tex,
	const unsigned col=GFX_white,const int m=10,const double texscale=1)
{
	glEnable(GL_TEXTURE_2D); glBindTexture(GL_TEXTURE_2D,tex);
	glBegin(GL_QUADS);
	spheretex_inner(ox,oy,oz,r,col,m,texscale);
	glEnd();
	glDisable(GL_TEXTURE_2D);
}

void cylinder_inner(const double x1,const double y1,const double z1,
	const double x2,const double y2,const double z2,
	const double r,const unsigned col,const int m=10)
{ // Put this inside glBegin(GL_QUADS), m is divisions in 180 degrees
	double ax=x2-x1,ay=y2-y1,az=z2-z1,bx,by,bz,cx,cy,cz,k,dx,dy,dz;
	k=ay*ay+az*az;
	if (k>1e-12)
	{ // b := a x (1,0,0)
		bx=0; by=az; bz=-ay;
	}
	else
	{ // b := a x (0,1,0)
		bx=-az; by=0; bz=ax;
		k=ax*ax+az*az;
	}
	k=1.0/sqrt(k); bx*=k; by*=k; bz*=k;
	// c := a x b
	cx=ay*bz-az*by; cy=az*bx-ax*bz; cz=ax*by-ay*bx;
	k=1.0/sqrt(cx*cx+cy*cy+cz*cz); cx*=k; cy*=k; cz*=k;
	int n; double *c=(double *)malloc((m+1)*sizeof(double)),*s=(double *)malloc((m+1)*sizeof(double)),th;
	for (n=m;n>=0;n--) {th=M_PI*n/m; c[n]=cos(th); s[n]=sin(th);}
	glColor3ub(col>>16,(col>>8)&0xFF,col&0xFF);
	for (n=m-1;n>=0;n--)
	{
		glNormal3d(dx=c[n]*bx+s[n]*cx,dy=c[n]*by+s[n]*cy,dz=c[n]*bz+s[n]*cz);
		glVertex3d(x2+r*dx,y2+r*dy,z2+r*dz);
		glNormal3d(dx,dy,dz);
		glVertex3d(x1+r*dx,y1+r*dy,z1+r*dz);
		glNormal3d(dx=c[n+1]*bx+s[n+1]*cx,dy=c[n+1]*by+s[n+1]*cy,dz=c[n+1]*bz+s[n+1]*cz);
		glVertex3d(x1+r*dx,y1+r*dy,z1+r*dz);
		glNormal3d(dx,dy,dz);
		glVertex3d(x2+r*dx,y2+r*dy,z2+r*dz);
		glNormal3d(dx=-c[n]*bx-s[n]*cx,dy=-c[n]*by-s[n]*cy,dz=-c[n]*bz-s[n]*cz);
		glVertex3d(x2+r*dx,y2+r*dy,z2+r*dz);
		glNormal3d(dx,dy,dz);
		glVertex3d(x1+r*dx,y1+r*dy,z1+r*dz);
		glNormal3d(dx=-c[n+1]*bx-s[n+1]*cx,dy=-c[n+1]*by-s[n+1]*cy,dz=-c[n+1]*bz-s[n+1]*cz);
		glVertex3d(x1+r*dx,y1+r*dy,z1+r*dz);
		glNormal3d(dx,dy,dz);
		glVertex3d(x2+r*dx,y2+r*dy,z2+r*dz);
	}
	free(c); free(s);
}

void cylinder(const double x1,const double y1,const double z1,
	const double x2,const double y2,const double z2,
	const double r,const unsigned col,const int m=10)
{ // Includes the ends (but not fast)
	double ax=x2-x1,ay=y2-y1,az=z2-z1,bx,by,bz,cx,cy,cz,k;
	k=ay*ay+az*az;
	if (k>1e-12)
	{ // b := a x (1,0,0)
		bx=0; by=az; bz=-ay;
	}
	else
	{ // b := a x (0,1,0)
		bx=-az; by=0; bz=ax;
		k=ax*ax+az*az;
	}
	k=1.0/sqrt(k); bx*=k; by*=k; bz*=k;
	// c := a x b
	cx=ay*bz-az*by; cy=az*bx-ax*bz; cz=ax*by-ay*bx;
	k=1.0/sqrt(cx*cx+cy*cy+cz*cz); cx*=k; cy*=k; cz*=k;
	k=1.0/sqrt(ax*ax+ay*ay+az*az); ax*=k; ay*=k; az*=k;
	int n,i; double *c=(double *)malloc((m*2+1)*sizeof(double)),*s=(double *)malloc((m*2+1)*sizeof(double)),th,x,y,z;
	for (n=m*2;n>=0;n--) {th=M_PI*n/m; c[n]=r*cos(th); s[n]=r*sin(th);}
	glBegin(GL_TRIANGLES);
	glColor3ub(col>>16,(col>>8)&0xFF,col&0xFF);
	for (i=1;i>=0;i--)
	{
		if (i) {glNormal3d(ax,ay,az); x=x2; y=y2; z=z2;}
		else {glNormal3d(-ax,-ay,-az); x=x1; y=y1; z=z1;}
		for (n=m*2-1;n>=0;n--)
		{
			glVertex3d(x,y,z);
			glVertex3d(x+c[n]*bx+s[n]*cx,y+c[n]*by+s[n]*cy,z+c[n]*bz+s[n]*cz);
			glVertex3d(x+c[n+1]*bx+s[n+1]*cx,y+c[n+1]*by+s[n+1]*cy,z+c[n+1]*bz+s[n+1]*cz);
		}
	}
	glEnd();
	free(c); free(s);
	glBegin(GL_QUADS);
	cylinder_inner(x1,y1,z1,x2,y2,z2,r,col,m);
	glEnd();
}

void tube(const double x1,const double y1,const double z1,
	const double x2,const double y2,const double z2,
	const double r1,const double r2,const unsigned col,const int m=10)
{ // Includes the ends (but not fast)
	double ax=x2-x1,ay=y2-y1,az=z2-z1,bx,by,bz,cx,cy,cz,k;
	k=ay*ay+az*az;
	if (k>1e-12)
	{ // b := a x (1,0,0)
		bx=0; by=az; bz=-ay;
	}
	else
	{ // b := a x (0,1,0)
		bx=-az; by=0; bz=ax;
		k=ax*ax+az*az;
	}
	k=1.0/sqrt(k); bx*=k; by*=k; bz*=k;
	// c := a x b
	cx=ay*bz-az*by; cy=az*bx-ax*bz; cz=ax*by-ay*bx;
	k=1.0/sqrt(cx*cx+cy*cy+cz*cz); cx*=k; cy*=k; cz*=k;
	k=1.0/sqrt(ax*ax+ay*ay+az*az); ax*=k; ay*=k; az*=k;
	int n; double *c=(double *)malloc((m+1)*sizeof(double)),*s=(double *)malloc((m+1)*sizeof(double)),th;
	for (n=m;n>=0;n--) {th=M_PI*n/m; c[n]=cos(th); s[n]=sin(th);}
	glBegin(GL_QUADS);
	glColor3ub(col>>16,(col>>8)&0xFF,col&0xFF);
	glNormal3d(-ax,-ay,-az);
	for (n=m-1;n>=0;n--)
	{
		glVertex3d(x1+r1*(c[n]*bx+s[n]*cx),y1+r1*(c[n]*by+s[n]*cy),z1+r1*(c[n]*bz+s[n]*cz));
		glVertex3d(x1+r1*(c[n+1]*bx+s[n+1]*cx),y1+r1*(c[n+1]*by+s[n+1]*cy),z1+r1*(c[n+1]*bz+s[n+1]*cz));
		glVertex3d(x1+r2*(c[n+1]*bx+s[n+1]*cx),y1+r2*(c[n+1]*by+s[n+1]*cy),z1+r2*(c[n+1]*bz+s[n+1]*cz));
		glVertex3d(x1+r2*(c[n]*bx+s[n]*cx),y1+r2*(c[n]*by+s[n]*cy),z1+r2*(c[n]*bz+s[n]*cz));
		glVertex3d(x1-r1*(c[n]*bx+s[n]*cx),y1-r1*(c[n]*by+s[n]*cy),z1-r1*(c[n]*bz+s[n]*cz));
		glVertex3d(x1-r1*(c[n+1]*bx+s[n+1]*cx),y1-r1*(c[n+1]*by+s[n+1]*cy),z1-r1*(c[n+1]*bz+s[n+1]*cz));
		glVertex3d(x1-r2*(c[n+1]*bx+s[n+1]*cx),y1-r2*(c[n+1]*by+s[n+1]*cy),z1-r2*(c[n+1]*bz+s[n+1]*cz));
		glVertex3d(x1-r2*(c[n]*bx+s[n]*cx),y1-r2*(c[n]*by+s[n]*cy),z1-r2*(c[n]*bz+s[n]*cz));
	}
	glNormal3d(ax,ay,az);
	for (n=m-1;n>=0;n--)
	{
		glVertex3d(x2+r1*(c[n]*bx+s[n]*cx),y2+r1*(c[n]*by+s[n]*cy),z2+r1*(c[n]*bz+s[n]*cz));
		glVertex3d(x2+r1*(c[n+1]*bx+s[n+1]*cx),y2+r1*(c[n+1]*by+s[n+1]*cy),z2+r1*(c[n+1]*bz+s[n+1]*cz));
		glVertex3d(x2+r2*(c[n+1]*bx+s[n+1]*cx),y2+r2*(c[n+1]*by+s[n+1]*cy),z2+r2*(c[n+1]*bz+s[n+1]*cz));
		glVertex3d(x2+r2*(c[n]*bx+s[n]*cx),y2+r2*(c[n]*by+s[n]*cy),z2+r2*(c[n]*bz+s[n]*cz));
		glVertex3d(x2-r1*(c[n]*bx+s[n]*cx),y2-r1*(c[n]*by+s[n]*cy),z2-r1*(c[n]*bz+s[n]*cz));
		glVertex3d(x2-r1*(c[n+1]*bx+s[n+1]*cx),y2-r1*(c[n+1]*by+s[n+1]*cy),z2-r1*(c[n+1]*bz+s[n+1]*cz));
		glVertex3d(x2-r2*(c[n+1]*bx+s[n+1]*cx),y2-r2*(c[n+1]*by+s[n+1]*cy),z2-r2*(c[n+1]*bz+s[n+1]*cz));
		glVertex3d(x2-r2*(c[n]*bx+s[n]*cx),y2-r2*(c[n]*by+s[n]*cy),z2-r2*(c[n]*bz+s[n]*cz));
	}
	free(c); free(s);
	cylinder_inner(x1,y1,z1,x2,y2,z2,-r1,col,m);
	cylinder_inner(x1,y1,z1,x2,y2,z2,r2,col,m);
	glEnd();
}

void cube_inner(const double x,const double y,const double z,const double s)
{ // Must be within glBegin(GL_QUADS)...
	glNormal3f(0,0,-1);
	glVertex3d(x,y,z); glVertex3d(x+s,y,z); glVertex3d(x+s,y+s,z); glVertex3d(x,y+s,z);
	glNormal3f(0,0,1);
	glVertex3d(x,y,z+s); glVertex3d(x+s,y,z+s); glVertex3d(x+s,y+s,z+s); glVertex3d(x,y+s,z+s);
	glNormal3f(0,-1,0);
	glVertex3d(x,y,z); glVertex3d(x+s,y,z); glVertex3d(x+s,y,z+s); glVertex3d(x,y,z+s);
	glNormal3f(0,1,0);
	glVertex3d(x,y+s,z); glVertex3d(x+s,y+s,z); glVertex3d(x+s,y+s,z+s); glVertex3d(x,y+s,z+s);
	glNormal3f(-1,0,0);
	glVertex3d(x,y,z); glVertex3d(x,y+s,z); glVertex3d(x,y+s,z+s); glVertex3d(x,y,z+s);
	glNormal3f(1,0,0);
	glVertex3d(x+s,y,z); glVertex3d(x+s,y+s,z); glVertex3d(x+s,y+s,z+s); glVertex3d(x+s,y,z+s);
}

void cubetex_inner(const double x,const double y,const double z,const double s)
{ // Also needs 2D texturing (one texture per cube face)
	glNormal3f(0,0,-1);
	glTexCoord2f(0,0); glVertex3d(x,y,z); glTexCoord2f(1,0); glVertex3d(x+s,y,z); glTexCoord2f(1,1); glVertex3d(x+s,y+s,z); glTexCoord2f(0,1); glVertex3d(x,y+s,z);
	glNormal3f(0,0,1);
	glTexCoord2f(0,0); glVertex3d(x,y,z+s); glTexCoord2f(1,0); glVertex3d(x+s,y,z+s); glTexCoord2f(1,1); glVertex3d(x+s,y+s,z+s); glTexCoord2f(0,1); glVertex3d(x,y+s,z+s);
	glNormal3f(0,-1,0);
	glTexCoord2f(0,0); glVertex3d(x,y,z); glTexCoord2f(1,0); glVertex3d(x+s,y,z); glTexCoord2f(1,1); glVertex3d(x+s,y,z+s); glTexCoord2f(0,1); glVertex3d(x,y,z+s);
	glNormal3f(0,1,0);
	glTexCoord2f(0,0); glVertex3d(x,y+s,z); glTexCoord2f(1,0); glVertex3d(x+s,y+s,z); glTexCoord2f(1,1); glVertex3d(x+s,y+s,z+s); glTexCoord2f(0,1); glVertex3d(x,y+s,z+s);
	glNormal3f(-1,0,0);
	glTexCoord2f(0,0); glVertex3d(x,y,z); glTexCoord2f(1,0); glVertex3d(x,y+s,z); glTexCoord2f(1,1); glVertex3d(x,y+s,z+s); glTexCoord2f(0,1); glVertex3d(x,y,z+s);
	glNormal3f(1,0,0);
	glTexCoord2f(0,0); glVertex3d(x+s,y,z); glTexCoord2f(1,0); glVertex3d(x+s,y+s,z); glTexCoord2f(1,1); glVertex3d(x+s,y+s,z+s); glTexCoord2f(0,1); glVertex3d(x+s,y,z+s);
}

void cuboid_inner(const double x0,const double x1,const double y0,const double y1,const double z0,const double z1)
{ // Must be within glBegin(GL_QUADS)...
	glNormal3f(0,0,-1);
	glVertex3d(x0,y0,z0); glVertex3d(x1,y0,z0); glVertex3d(x1,y1,z0); glVertex3d(x0,y1,z0);
	glNormal3f(0,0,1);
	glVertex3d(x0,y0,z1); glVertex3d(x1,y0,z1); glVertex3d(x1,y1,z1); glVertex3d(x0,y1,z1);
	glNormal3f(0,-1,0);
	glVertex3d(x0,y0,z0); glVertex3d(x1,y0,z0); glVertex3d(x1,y0,z1); glVertex3d(x0,y0,z1);
	glNormal3f(0,1,0);
	glVertex3d(x0,y1,z0); glVertex3d(x1,y1,z0); glVertex3d(x1,y1,z1); glVertex3d(x0,y1,z1);
	glNormal3f(-1,0,0);
	glVertex3d(x0,y0,z0); glVertex3d(x0,y1,z0); glVertex3d(x0,y1,z1); glVertex3d(x0,y0,z1);
	glNormal3f(1,0,0);
	glVertex3d(x1,y0,z0); glVertex3d(x1,y1,z0); glVertex3d(x1,y1,z1); glVertex3d(x1,y0,z1);
}

void cuboidtex_inner(const double x0,const double x1,const double y0,const double y1,const double z0,const double z1,const double tsc)
{ // Also needs 2D texturing (texture repeats at tsc in size)
	double tx=(x1-x0)/tsc,ty=(y1-y0)/tsc,tz=(z1-z0)/tsc;
	glNormal3f(0,0,-1);
	glTexCoord2d(0,0); glVertex3d(x0,y0,z0); glTexCoord2d(tx,0); glVertex3d(x1,y0,z0); glTexCoord2d(tx,ty); glVertex3d(x1,y1,z0); glTexCoord2d(0,ty); glVertex3d(x0,y1,z0);
	glNormal3f(0,0,1);
	glTexCoord2d(0,0); glVertex3d(x0,y0,z1); glTexCoord2d(tx,0); glVertex3d(x1,y0,z1); glTexCoord2d(tx,ty); glVertex3d(x1,y1,z1); glTexCoord2d(0,ty); glVertex3d(x0,y1,z1);
	glNormal3f(0,-1,0);
	glTexCoord2d(0,0); glVertex3d(x0,y0,z0); glTexCoord2d(tx,0); glVertex3d(x1,y0,z0); glTexCoord2d(tx,tz); glVertex3d(x1,y0,z1); glTexCoord2d(0,tz); glVertex3d(x0,y0,z1);
	glNormal3f(0,1,0);
	glTexCoord2d(0,0); glVertex3d(x0,y1,z0); glTexCoord2d(tx,0); glVertex3d(x1,y1,z0); glTexCoord2d(tx,tz); glVertex3d(x1,y1,z1); glTexCoord2d(0,tz); glVertex3d(x0,y1,z1);
	glNormal3f(-1,0,0);
	glTexCoord2d(0,0); glVertex3d(x0,y0,z0); glTexCoord2d(ty,0); glVertex3d(x0,y1,z0); glTexCoord2d(ty,tz); glVertex3d(x0,y1,z1); glTexCoord2d(0,tz); glVertex3d(x0,y0,z1);
	glNormal3f(1,0,0);
	glTexCoord2d(0,0); glVertex3d(x1,y0,z0); glTexCoord2d(ty,0); glVertex3d(x1,y1,z0); glTexCoord2d(ty,tz); glVertex3d(x1,y1,z1); glTexCoord2d(0,tz); glVertex3d(x1,y0,z1);
}

void cuboid_wireframe_inner(const double x0,const double x1,const double y0,const double y1,const double z0,const double z1)
{ // Must be within glBegin(GL_LINES)
	glVertex3d(x0,y0,z0); glVertex3d(x1,y0,z0);
	glVertex3d(x0,y0,z1); glVertex3d(x1,y0,z1);
	glVertex3d(x0,y1,z0); glVertex3d(x1,y1,z0);
	glVertex3d(x0,y1,z1); glVertex3d(x1,y1,z1);
	glVertex3d(x0,y0,z0); glVertex3d(x0,y1,z0);
	glVertex3d(x0,y0,z1); glVertex3d(x0,y1,z1);
	glVertex3d(x1,y0,z0); glVertex3d(x1,y1,z0);
	glVertex3d(x1,y0,z1); glVertex3d(x1,y1,z1);
	glVertex3d(x0,y0,z0); glVertex3d(x0,y0,z1);
	glVertex3d(x0,y1,z0); glVertex3d(x0,y1,z1);
	glVertex3d(x1,y0,z0); glVertex3d(x1,y0,z1);
	glVertex3d(x1,y1,z0); glVertex3d(x1,y1,z1);
}

void sphere_octant_inner(const double ox,const double oy,const double oz,const double r,
	const int kx,const int ky,const int kz,
	const unsigned col,const int m)
{ // m is number of divisions in 180 degrees (must be even)
	int n,i; double th,
		*c=(double *)malloc((m/2+1)*sizeof(double)),
		*s=(double *)malloc((m/2+1)*sizeof(double));
	for (n=m/2;n>=0;n--) {th=M_PI*n/m; c[n]=cos(th); s[n]=sin(th);}
	glColor3ub(col>>16,(col>>8)&0xFF,col&0xFF);
	for (n=m/2-1;n>=0;n--) for (i=m/2-1;i>=0;i--)
	{
		glNormal3d(c[n]*s[i]*kx,s[n]*s[i]*ky,c[i]*kz);
		glVertex3d(ox+r*c[n]*s[i]*kx,oy+r*s[n]*s[i]*ky,oz+r*c[i]*kz);
		glNormal3d(c[n+1]*s[i]*kx,s[n+1]*s[i]*ky,c[i]*kz);
		glVertex3d(ox+r*c[n+1]*s[i]*kx,oy+r*s[n+1]*s[i]*ky,oz+r*c[i]*kz);
		glNormal3d(c[n+1]*s[i+1]*kx,s[n+1]*s[i+1]*ky,c[i+1]*kz);
		glVertex3d(ox+r*c[n+1]*s[i+1]*kx,oy+r*s[n+1]*s[i+1]*ky,oz+r*c[i+1]*kz);
		glNormal3d(c[n]*s[i+1]*kx,s[n]*s[i+1]*ky,c[i+1]*kz);
		glVertex3d(ox+r*c[n]*s[i+1]*kx,oy+r*s[n]*s[i+1]*ky,oz+r*c[i+1]*kz);
	}
	free(c); free(s);
}

// Advanced OpenGL graphics
void blur(float r)
{
	static GLuint scr=0; if (!scr) glGenTextures(1,&scr);
	GLboolean oldtex2d=glIsEnabled(GL_TEXTURE_2D),oldlight=glIsEnabled(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D); glDisable(GL_LIGHTING);
	glBindTexture(GL_TEXTURE_2D,scr);
	int xs,ys; for (xs=1;xs<xres;xs<<=1); for (ys=1;ys<yres;ys<<=1);
	GLfloat tx=(GLfloat)xres/xs,ty=(GLfloat)yres/ys,tr;
	ortho_on();
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	for (;r>=1;r*=0.5)
	{ /**/ // Would glCopyTexSubImage2D be faster here?
		glCopyTexImage2D(GL_TEXTURE_2D,0,GL_RGB8,0,0,xs,ys,0); // Not GL_RGBA8, that causes weird mallards
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glBegin(GL_QUADS);
		glColor4f(1,1,1,1.0/3);
		tr=(GLfloat)r/xs;
		glTexCoord2f(tr,ty); glVertex2f(0,0);
		glTexCoord2f(tx+tr,ty); glVertex2f(xres,0);
		glTexCoord2f(tx+tr,0); glVertex2f(xres,yres);
		glTexCoord2f(tr,0); glVertex2f(0,yres);
		glColor4f(1,1,1,1.0/4);
		glTexCoord2f(-tr,ty); glVertex2f(0,0);
		glTexCoord2f(tx-tr,ty); glVertex2f(xres,0);
		glTexCoord2f(tx-tr,0); glVertex2f(xres,yres);
		glTexCoord2f(-tr,0); glVertex2f(0,yres);
		/*glColor4f(1,1,1,1.0/3);
		glTexCoord2f(0,ty); glVertex2f(0,r);
		glTexCoord2f(tx,ty); glVertex2f(xres,r);
		glTexCoord2f(tx,0); glVertex2f(xres,yres+r);
		glTexCoord2f(0,0); glVertex2f(0,yres+r);
		glColor4f(1,1,1,1.0/4);
		glTexCoord2f(0,ty); glVertex2f(r,r);
		glTexCoord2f(tx,ty); glVertex2f(xres+r,r);
		glTexCoord2f(tx,0); glVertex2f(xres+r,yres+r);
		glTexCoord2f(0,0); glVertex2f(r,yres+r);*/
		glEnd();
		glCopyTexImage2D(GL_TEXTURE_2D,0,GL_RGB8,0,0,xs,ys,0); // Not GL_RGBA8, that causes weird mallards
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
		glBegin(GL_QUADS);
		glColor4f(1,1,1,1.0/3);
		tr=(GLfloat)r/ys;
		glTexCoord2f(0,ty+tr); glVertex2f(0,0);
		glTexCoord2f(tx,ty+tr); glVertex2f(xres,0);
		glTexCoord2f(tx,tr); glVertex2f(xres,yres);
		glTexCoord2f(0,tr); glVertex2f(0,yres);
		glColor4f(1,1,1,1.0/4);
		glTexCoord2f(0,ty-tr); glVertex2f(0,0);
		glTexCoord2f(tx,ty-tr); glVertex2f(xres,0);
		glTexCoord2f(tx,-tr); glVertex2f(xres,yres);
		glTexCoord2f(0,-tr); glVertex2f(0,yres);
		glEnd();
	}
	//glDeleteTextures(1,&scr); // Creating and deleting the texture object was the slow part!!
	ortho_off();
	glBlendFunc(GL_ONE,GL_ZERO);
	if (!oldtex2d) glDisable(GL_TEXTURE_2D);
	if (oldlight) glEnable(GL_LIGHTING);
}

#include <rnd/sjbbuild2dmipmaps.c>

GLuint makecubemap(void colfunc(double,double,double,unsigned char *),const unsigned res,const unsigned aa=1)
{ // colfunc accepts "direction cosines" and writes the RGB bytes starting at the supplied pointer, res should be power of two
	int x,y,i; GLuint ret; glGenTextures(1,&ret);
	glBindTexture(GL_TEXTURE_CUBE_MAP,ret);
	unsigned char *datapz=(unsigned char *)malloc(res*res*3),*datanz=(unsigned char *)malloc(res*res*3),
		*datapy=(unsigned char *)malloc(res*res*3),*datany=(unsigned char *)malloc(res*res*3),
		*datapx=(unsigned char *)malloc(res*res*3),*datanx=(unsigned char *)malloc(res*res*3);
	double fx,fy,k,rh=0.5*res;
	if (aa<=1) for (y=res-1;y>=0;y--) for (x=res-1;x>=0;x--)
	{
		fx=(x+0.5-rh)/rh; fy=(y+0.5-rh)/rh; k=1.0/sqrt(1.0+fx*fx+fy*fy);
		fx*=k; fy*=k; i=(y*res+x)*3;
		colfunc(fx,-fy,k,datapz+i);
		colfunc(-fx,-fy,-k,datanz+i);
		colfunc(fx,k,fy,datapy+i);
		colfunc(fx,-k,-fy,datany+i);
		colfunc(k,-fy,-fx,datapx+i);
		colfunc(-k,-fy,fx,datanx+i);
	}
	else // Antialiasing, pretty CPU-intensive!
	{
		int n,aas=aa*aa; unsigned char s[18],c; double ss[18],iaas=1.0/aas;
		for (y=res-1;y>=0;y--) for (x=res-1;x>=0;x--)
		{
			for (n=17;n>=0;n--) ss[n]=0;
			for (n=aas-1;n>=0;n--)
			{
				i=aa-1-n/aa+n%aa*aa;
				fx=(x-rh+(0.5+n)/aas)/rh; fy=(y-rh+(0.5+i)/aas)/rh;
				k=1.0/sqrt(1.0+fx*fx+fy*fy); fx*=k; fy*=k;
				colfunc(fx,-fy,k,s);
				colfunc(-fx,-fy,-k,s+3);
				colfunc(fx,k,fy,s+6);
				colfunc(fx,-k,-fy,s+9);
				colfunc(k,-fy,-fx,s+12);
				colfunc(-k,-fy,fx,s+15);
				for (i=17;i>=0;i--) ss[i]+=s[i];
			}
			i=(y*res+x)*3;
			for (n=2;n>=0;n--)
			{
				datapz[i+n]=c=(unsigned char)(iaas*ss[n]);
				datanz[i+n]=c=(unsigned char)(iaas*ss[n+3]);
				datapy[i+n]=c=(unsigned char)(iaas*ss[n+6]);
				datany[i+n]=c=(unsigned char)(iaas*ss[n+9]);
				datapx[i+n]=c=(unsigned char)(iaas*ss[n+12]);
				datanx[i+n]=c=(unsigned char)(iaas*ss[n+15]);
			}
		}
	}
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datapz);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datanz);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datapy);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datany);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datapx);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X,0,GL_RGB,res,res,0,GL_RGB,GL_UNSIGNED_BYTE,datanx);
	//sjbBuildCubeMipmaps(3,res,datapx,datanx,datapy,datany,datapz,datanz);
	glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR); // GL_LINEAR for glTexImage2D
	glTexParameteri(GL_TEXTURE_CUBE_MAP,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	aniso(GL_TEXTURE_CUBE_MAP);
	free(datapz); free(datanz); free(datapy); free(datany); free(datapx); free(datanx);
	return ret;
}

void squareproj_on(void)
{ // Render the side of a cube (Lighting_squareprojbegin from RNode/lighting.c)
	double m[4][4];
	glGetDoublev(GL_PROJECTION_MATRIX,(double *)m);
	//o.xtan=1.0/m[0][0]; o.ytan=1.0/m[1][1];
	m[0][0]=1; m[1][1]=1;
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadMatrixd((double *)m);
	glMatrixMode(GL_MODELVIEW);
}

void squareproj_off(void)
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void frame_lookfrom(const float *m)
{ // Places the eye where the object is
	float i[16];
	i[0]=m[0]; i[4]=m[1]; i[8]=m[2]; i[12]=-m[0]*m[12]-m[1]*m[13]-m[2]*m[14];
	i[1]=m[4]; i[5]=m[5]; i[9]=m[6]; i[13]=-m[4]*m[12]-m[5]*m[13]-m[6]*m[14];
	i[2]=m[8]; i[6]=m[9]; i[10]=m[10]; i[14]=-m[8]*m[12]-m[9]*m[13]-m[10]*m[14];
	i[3]=0; i[7]=0; i[11]=0; i[15]=1;
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(i);
}

GLuint scenecubemap(void scenefunc(void *),const double *o,const unsigned res,void *arg=NULL,
	const GLuint update=0) // Setting update!=0 updates an existing texture: always do this if possible, texture alloc can be the slowest thing in a loop
{ // Generates a cube map from a function that plots a scene (based on RNode_cubemap from RNode/lighting.c)
	int n,i; GLuint ret;
	if (!update) glGenTextures(1,&ret); else ret=update;
	glBindTexture(GL_TEXTURE_CUBE_MAP,ret);
	squareproj_on();
	glPushMatrix(); glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0,0,res,res);
	float m[16],ox=0,oy=0,oz=0;
	for (n=0;n<6;n++)
	{
		clearrect(0,0,res,res);
		for (i=11;i>=0;i--) m[i]=0;
		m[12]=o[0]; m[13]=o[1]; m[14]=o[2]; m[15]=1;
		switch (n)
		{
			case 0: m[2]=-1; m[5]=-1; m[8]=-1; break;
			case 1: m[2]=1; m[5]=-1; m[8]=1; break;
			case 2: m[0]=1; m[6]=1; m[9]=-1; break;
			case 3: m[0]=1; m[6]=-1; m[9]=1;  break;
			case 4: m[0]=1; m[5]=-1; m[10]=-1; break;
			case 5: m[0]=-1; m[5]=-1; m[10]=1; break;
		}
		frame_lookfrom(m);
		scenefunc(arg);
		glBindTexture(GL_TEXTURE_CUBE_MAP,ret); // Re-bind in case other cubemaps got used in scenefunc
		switch (n)
		{
			case 0: i=GL_TEXTURE_CUBE_MAP_POSITIVE_X; break; case 1: i=GL_TEXTURE_CUBE_MAP_NEGATIVE_X; break;
			case 2: i=GL_TEXTURE_CUBE_MAP_POSITIVE_Y; break; case 3: i=GL_TEXTURE_CUBE_MAP_NEGATIVE_Y; break;
			case 4: i=GL_TEXTURE_CUBE_MAP_POSITIVE_Z; break; case 5: i=GL_TEXTURE_CUBE_MAP_NEGATIVE_Z; break;
		}
		if (update) glCopyTexSubImage2D(i,0, 0,0, 0,0,res,res);
		else glCopyTexImage2D(i,0,GL_RGB8, 0,0,res,res, 0);
	}
	clearrect(0,0,res,res); // So depth values don't interfere with plotting
	glPopAttrib(); glPopMatrix();
	squareproj_off();
	glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	aniso(GL_TEXTURE_CUBE_MAP);
	return ret;
}

void plotcubemap_inner(void)
{
	glBegin(GL_QUADS);
	glTexCoord3f(1,1,1); glVertex3f(1,1,1);
	glTexCoord3f(1,1,-1); glVertex3f(1,1,-1);
	glTexCoord3f(1,-1,-1); glVertex3f(1,-1,-1);
	glTexCoord3f(1,-1,1); glVertex3f(1,-1,1);
	glTexCoord3f(1,1,1); glVertex3f(1,1,1);
	glTexCoord3f(1,1,-1); glVertex3f(1,1,-1);
	glTexCoord3f(-1,1,-1); glVertex3f(-1,1,-1);
	glTexCoord3f(-1,1,1); glVertex3f(-1,1,1);
	glTexCoord3f(1,1,1); glVertex3f(1,1,1);
	glTexCoord3f(1,-1,1); glVertex3f(1,-1,1);
	glTexCoord3f(-1,-1,1); glVertex3f(-1,-1,1);
	glTexCoord3f(-1,1,1); glVertex3f(-1,1,1);
	glTexCoord3f(-1,-1,-1); glVertex3f(-1,-1,-1);
	glTexCoord3f(-1,-1,1); glVertex3f(-1,-1,1);
	glTexCoord3f(-1,1,1); glVertex3f(-1,1,1);
	glTexCoord3f(-1,1,-1); glVertex3f(-1,1,-1);
	glTexCoord3f(-1,-1,-1); glVertex3f(-1,-1,-1);
	glTexCoord3f(-1,-1,1); glVertex3f(-1,-1,1);
	glTexCoord3f(1,-1,1); glVertex3f(1,-1,1);
	glTexCoord3f(1,-1,-1); glVertex3f(1,-1,-1);
	glTexCoord3f(-1,-1,-1); glVertex3f(-1,-1,-1);
	glTexCoord3f(-1,1,-1); glVertex3f(-1,1,-1);
	glTexCoord3f(1,1,-1); glVertex3f(1,1,-1);
	glTexCoord3f(1,-1,-1); glVertex3f(1,-1,-1);
	glEnd();
}

void plotcubemap(const float ox,const float oy,const float oz,const float r,const GLuint cubemap)
{
	glEnable(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP,cubemap);
	glPushMatrix(); glTranslatef(ox,oy,oz); glScalef(r,r,r);
	glColor3f(1,1,1);
	plotcubemap_inner();
	glPopMatrix();
	glDisable(GL_TEXTURE_CUBE_MAP);
}

void skysphere(const GLuint cubemap)
{ // Sky sphere; put the inverse of the viewer rotation matrix in the GL_TEXTURE matrix first! (similar for cubemaps/reflection maps)
	glPushAttrib(GL_ENABLE_BIT|GL_TEXTURE_BIT);
	glPushMatrix(); glLoadIdentity();
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP,cubemap);
	glColor3f(1,1,1);
	plotcubemap_inner();
	glClear(GL_DEPTH_BUFFER_BIT);
	glPopMatrix();
	glPopAttrib();
}

#define INCLUDED_SBCLIB_GRAPHICS_OPENGL
#endif
